/*
LUFA Library
Copyright (C) Dean Camera, 2009.

dean [at] fourwalledcubicle [dot] com
www.fourwalledcubicle.com
*/

/*
Copyright 2009  Dean Camera (dean [at] fourwalledcubicle [dot] com)

Permission to use, copy, modify, and distribute this software
and its documentation for any purpose and without fee is hereby
granted, provided that the above copyright notice appear in all
copies and that both that the copyright notice and this
permission notice and warranty disclaimer appear in supporting
documentation, and that the name of the author not be used in
advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

The author disclaim all warranties with regard to this
software, including all implied warranties of merchantability
and fitness.  In no event shall the author be liable for any
special, indirect or consequential damages or any damages
whatsoever resulting from loss of use, data or profits, whether
in an action of contract, negligence or other tortious action,
arising out of or in connection with the use or performance of
this software.
*/

#include "Orbee.h"
#include <util/delay.h>


static void orb_init( void );

// at90usb1282 
// void (*start_bootloader) (void)=(void (*)(void))0xf000;
// at90usb162 
void (*start_bootloader) (void)=(void (*)(void))0x3000;


/* Scheduler Task List */
TASK_LIST
{
    { .Task = USB_USBTask          , .TaskStatus = TASK_STOP },
    { .Task = CDC_Task             , .TaskStatus = TASK_STOP },
};

/* Globals: */
/** Contains the current baud rate and other settings of the virtual serial port.
*
*  These values are set by the host via a class-specific request, and the physical USART should be reconfigured to match the
*  new settings each time they are changed by the host.
*/
CDC_Line_Coding_t LineCoding = { .BaudRateBPS = 9600,
.CharFormat  = OneStopBit,
.ParityType  = Parity_None,
.DataBits    = 8            };

/** Main program entry point. This routine configures the hardware required by the application, then
*  starts the scheduler to run the application tasks.
*/
int main(void)
{
    /* Disable watchdog if enabled by bootloader/fuses */
    MCUSR &= ~(1 << WDRF);
    wdt_disable();

    /* Disable clock division */
    clock_prescale_set(clock_div_1);

    /* Hardware Initialization */
    orb_init();

    /* Indicate USB not ready */
    //UpdateStatus(Status_USBNotReady);

    /* Initialize Scheduler so that it can be used */
    Scheduler_Init();

    /* Initialize USB Subsystem */
    USB_Init();

    /* Scheduling - routine never returns, so put this last in the main function */
    Scheduler_Start();
}

/** Event handler for the USB_Connect event. This indicates that the device is enumerating via the status LEDs and
*  starts the library USB task to begin the enumeration and USB management process.
*/
void EVENT_USB_Connect(void)
{
    /* Start USB management task */
    Scheduler_SetTaskMode(USB_USBTask, TASK_RUN);

    /* Indicate USB enumerating */
    //UpdateStatus(Status_USBEnumerating);
}

/** Event handler for the USB_Disconnect event. This indicates that the device is no longer connected to a host via
*  the status LEDs and stops the USB management and CDC management tasks.
*/
void EVENT_USB_Disconnect(void)
{
    /* Stop running CDC and USB management tasks */
    Scheduler_SetTaskMode(CDC_Task, TASK_STOP);
    Scheduler_SetTaskMode(USB_USBTask, TASK_STOP);

    /* Indicate USB not ready */
    //UpdateStatus(Status_USBNotReady);
}

/** Event handler for the USB_ConfigurationChanged event. This is fired when the host set the current configuration
*  of the USB device after enumeration - the device endpoints are configured and the CDC management task started.
*/
void EVENT_USB_ConfigurationChanged(void)
{
    /* Setup CDC Notification, Rx and Tx Endpoints */
    Endpoint_ConfigureEndpoint(CDC_NOTIFICATION_EPNUM, EP_TYPE_INTERRUPT,
        ENDPOINT_DIR_IN, CDC_NOTIFICATION_EPSIZE,
        ENDPOINT_BANK_SINGLE);

    Endpoint_ConfigureEndpoint(CDC_TX_EPNUM, EP_TYPE_BULK,
        ENDPOINT_DIR_IN, CDC_TXRX_EPSIZE,
        ENDPOINT_BANK_SINGLE);

    Endpoint_ConfigureEndpoint(CDC_RX_EPNUM, EP_TYPE_BULK,
        ENDPOINT_DIR_OUT, CDC_TXRX_EPSIZE,
        ENDPOINT_BANK_SINGLE);

    /* Indicate USB connected and ready */
    //UpdateStatus(Status_USBReady);

    /* Start CDC task */
    Scheduler_SetTaskMode(CDC_Task, TASK_RUN);
}

/** Event handler for the USB_UnhandledControlPacket event. This is used to catch standard and class specific
*  control requests that are not handled internally by the USB library (including the CDC control commands,
*  which are all issued via the control endpoint), so that they can be handled appropriately for the application.
*/
void EVENT_USB_UnhandledControlPacket(void)
{
    uint8_t* LineCodingData = (uint8_t*)&LineCoding;

    /* Process CDC specific control requests */
    switch (USB_ControlRequest.bRequest)
    {
    case REQ_GetLineEncoding:
        if (USB_ControlRequest.bmRequestType == (REQDIR_DEVICETOHOST | REQTYPE_CLASS | REQREC_INTERFACE))
        {	
            /* Acknowledge the SETUP packet, ready for data transfer */
            Endpoint_ClearSETUP();

            /* Write the line coding data to the control endpoint */
            Endpoint_Write_Control_Stream_LE(LineCodingData, sizeof(LineCoding));

            /* Finalize the stream transfer to send the last packet or clear the host abort */
            Endpoint_ClearOUT();
        }

        break;
    case REQ_SetLineEncoding:
        if (USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
        {
            /* Acknowledge the SETUP packet, ready for data transfer */
            Endpoint_ClearSETUP();

            /* Read the line coding data in from the host into the global struct */
            Endpoint_Read_Control_Stream_LE(LineCodingData, sizeof(LineCoding));

            /* Finalize the stream transfer to clear the last packet from the host */
            Endpoint_ClearIN();
        }

        break;
    case REQ_SetControlLineState:
        if (USB_ControlRequest.bmRequestType == (REQDIR_HOSTTODEVICE | REQTYPE_CLASS | REQREC_INTERFACE))
        {				
            /* Acknowledge the SETUP packet, ready for data transfer */
            Endpoint_ClearSETUP();

            /* NOTE: Here you can read in the line state mask from the host, to get the current state of the output handshake
            lines. The mask is read in from the wValue parameter in USB_ControlRequest, and can be masked against the
            CONTROL_LINE_OUT_* masks to determine the RTS and DTR line states using the following code:
            */

            /* Acknowledge status stage */
            while (!(Endpoint_IsINReady()));
            Endpoint_ClearIN();
        }

        break;
    }
}


/** Function to manage status updates to the user. This is done via LEDs on the given board, if available, but may be changed to
*  log to a serial port, or anything else that is suitable for status updates.
*
*  \param CurrentStatus  Current status of the system, from the USBtoSerial_StatusCodes_t enum
*/
//void UpdateStatus(uint8_t CurrentStatus)
//{
//    uint8_t LEDMask = LEDS_NO_LEDS;
//
//    /* Set the LED mask to the appropriate LED mask based on the given status code */
//    switch (CurrentStatus)
//    {
//    case Status_USBNotReady:
//        LEDMask = (LEDS_LED1);
//        break;
//    case Status_USBEnumerating:
//        LEDMask = (LEDS_LED1 | LEDS_LED2);
//        break;
//    case Status_USBReady:
//        LEDMask = (LEDS_LED2 | LEDS_LED4);
//        break;
//    }
//
//    /* Set the board LEDs to the new LED mask */
//    LEDs_SetAllLEDs(LEDMask);
//}

































static void InitLights( void );



#define MSG_START 'C'
#define SEPARATOR ':'
#define EOL '\n'

struct LightCmd
{
    uint8_t R;
    uint8_t G;
    uint8_t B;
    uint16_t T;
};

#define MAX_COMMANDS 40
static struct LightCmd NextCmd[MAX_COMMANDS];
static struct LightCmd CurrentCmd[MAX_COMMANDS];
static uint8_t nCommands;
static uint8_t nCurrentCmd;
static uint16_t nCmdTimer;
static uint8_t fromR;
static uint8_t fromG;
static uint8_t fromB;
static uint8_t toR;
static uint8_t toG;
static uint8_t toB;
static uint16_t nPeriod;

static uint8_t nReadState;
static uint8_t nReadColourByte;
static uint8_t nColourByte;
static uint8_t nCmdPart;
static uint16_t nTimePart;
static bool bSendAck;

enum InputStateTable {
    WAIT_FOR_START,
    READ_BOOTLOADER,
    READ_COLOUR,
    READ_COLOUR_SEPARATOR,
    READ_TIME
};


static void orb_init( void )
{
    DDRC |= _BV(PC5) | _BV(PC6);
    DDRB |= _BV(PB7);

    // all compare values off
    OCR1A = 0;
    OCR1B = 0;
    OCR1C = 0;

    TCCR1A = _BV(WGM10) | _BV(COM1A1) | _BV(COM1B1) | _BV(COM1C1);              // clear OC on compare match
    TCCR1B = _BV(CS11) | _BV(CS10);     // clk/64 -> with Phase Correct PWM 0xFF, 244Hz

    InitLights();
    nReadState = WAIT_FOR_START;
}


static uint8_t FromHex( uint8_t asc )
{
    if ( asc >= '0' && asc <= '9' )
        return asc - '0';
    asc |= 0x20;    // to uppercase
    if ( asc >= 'a' && asc <= 'f' )
        return asc - 'a' + 10;
    return 0;
}

#define countof(x)  ((sizeof(x))/(sizeof((x)[0])))

static void SetLights( uint8_t R, uint8_t G, uint8_t B )
{
    OCR1AL = G; 
    OCR1BL = R; 
    OCR1CL = B; 
}

static void SetNextColour( void )
{
    nCmdTimer = 0;
    nCurrentCmd++;
    if ( nCurrentCmd >= nCommands )
        nCurrentCmd = 0;

    int8_t nNextCmd = nCurrentCmd + 1;
    if ( nNextCmd >= nCommands )
        nNextCmd = 0;

    fromR = CurrentCmd[nCurrentCmd].R;
    fromG = CurrentCmd[nCurrentCmd].G;
    fromB = CurrentCmd[nCurrentCmd].B;

    toR = CurrentCmd[nNextCmd].R;
    toG = CurrentCmd[nNextCmd].G;
    toB = CurrentCmd[nNextCmd].B;

    nPeriod = CurrentCmd[nCurrentCmd].T;
}

static void SetPattern( struct LightCmd *cmd, uint8_t count )
{
    memcpy( CurrentCmd, cmd, sizeof(struct LightCmd) * count );
    nCommands = count;
    nCurrentCmd = -1;
    SetNextColour();
}

static void InitLights( void )
{
    // Read from eeprom.
    // If nothing there, default.
    {
        static struct LightCmd default_cmd[] = { { 0x80, 0x80, 0x80, 1000 }, { 0, 0, 0, 1000 } };
        SetPattern( default_cmd, countof(default_cmd) );
    }
}

static uint8_t interpolate( uint8_t from, uint8_t to, uint16_t timer, uint16_t period )
{
    if ( period == 0 )
        return to;
    else
    {
        int32_t t = to - from;
        t *= timer;
        t /= period;
        t += from;
        return (uint8_t)(t & 0xFF);
    }
}

static void ProcessByte( uint8_t b )
{
    static char BootLoader[] = {'O','O','T','L','O','A','D','E','R'};

    switch ( nReadState )
    {
    case WAIT_FOR_START:
        if ( b == MSG_START )
        {
            nReadState = READ_COLOUR;
            nReadColourByte = 0;
            nColourByte = 0;
            nCmdPart = 0;
            memset( NextCmd, sizeof(NextCmd), 0 );
        }
        else if ( b == 'B' )
        {
            nReadState = READ_BOOTLOADER;
            nCmdPart = 0;
        }
        break;
    case READ_BOOTLOADER:
        if ( BootLoader[nCmdPart] == b ) 
        {
            nCmdPart++;
            if ( nCmdPart == countof(BootLoader) )
            {
		// disable interrupts we may have started
		TIMSK0 = 0;	// timers
		TIMSK1 = 0;
		UCSR1B = 0;	// usart
		EIMSK = 0;	// external interrupts
		PCICR = 0;

		// Reset ports
		DDRB = 0; PORTB = 0;
		DDRC = 0; PORTC = 0;
		DDRD = 0; PORTD = 0;

		USB_ShutDown();

                _delay_ms( 500 );		// 1/2sec delay
    
		//cli();
		//CLKPR = (1 << CLKPCE);
		//CLKPR = bPreScaler;
		//sei();

		_delay_us( 500 );

		(*start_bootloader)();           //! Jumping to booltoader
            }
        }
        else
            nReadState = WAIT_FOR_START;
        break;
    case READ_COLOUR:
        if ( (b >= '0' && b <= '9') ||
            (b >= 'a' && b <= 'f') ||
            (b >= 'A' && b <= 'F') )
        {
            uint8_t n = FromHex( b );
            if ( nReadColourByte & 1 )  // second nibble
                nColourByte |= n;
            else
            {
                nColourByte = n << 4;
                *((uint8_t *)(&NextCmd[nCmdPart]) + (nReadColourByte>>1)) = nColourByte;
            }
            nReadColourByte++;
            if ( nReadColourByte >= 6 )
            {
                nReadState = READ_COLOUR_SEPARATOR;
            }
        }
        else
            nReadState = WAIT_FOR_START;    // unknown character
        break;
    case READ_COLOUR_SEPARATOR:
        if ( b == ':' )
        {
            nTimePart = 0;
            nReadState = READ_TIME;
        }
        else if ( b == '\n' || b == '\r' )  // time is optional
        {
            nCmdPart++;                         
            SetPattern( NextCmd, nCmdPart );
            bSendAck = true;
            nReadState = WAIT_FOR_START;    
        }
        else
            nReadState = WAIT_FOR_START;    // Error
        break;
    case READ_TIME:
        if ( b == ':' || b == '\n' || b=='\r' )
        {
            NextCmd[nCmdPart].T = nTimePart/4;
            nCmdPart++;
            if ( b == ':' && nCmdPart < MAX_COMMANDS )  // don't take more than we can hold
            {
                nReadState = READ_COLOUR;
                nReadColourByte = 0;
                nColourByte = 0;
            }
            else
            {
                SetPattern( NextCmd, nCmdPart );
                bSendAck = true;
                nReadState = WAIT_FOR_START;    // Done
            }
            break;
        }
        else if (b >= '0' && b <= '9') 
        {
            uint8_t n = b - '0';
            nTimePart = (nTimePart * 10) + n;
        }
        else
            nReadState = WAIT_FOR_START;    // Error
        break;
    }
}


TASK(CDC_Task)
{
    if (USB_IsConnected)
    {
        /* Select the Serial Rx Endpoint */
        Endpoint_SelectEndpoint(CDC_RX_EPNUM);

        /* Check to see if a packet has been received from the host */
        if (Endpoint_IsOUTReceived())
        {
            /* Process the bytes */
            while ( Endpoint_BytesInEndpoint() )
            {
                uint8_t b = Endpoint_Read_Byte();
                ProcessByte( b );
            }

            if (!(Endpoint_BytesInEndpoint()))
                Endpoint_ClearOUT();
        }

        if ( bSendAck )
        {
            Endpoint_SelectEndpoint(CDC_TX_EPNUM);

            if ( Endpoint_IsReadWriteAllowed() )
            {
                Endpoint_Write_Stream_LE("OK\r\n", 4);
                /* Send the data */
                Endpoint_ClearIN();
                bSendAck = false;
            }
        }
    }


    // Do the colour cycling
    if ( TIFR1 & _BV(TOV1) )    // Timer overflow @ 244Hz
    {
        TIFR1 |= _BV(TOV1);     // clear overflow flag
        nCmdTimer++;     

        if ( nCmdTimer >= nPeriod )
            SetNextColour();

        uint8_t R = interpolate( fromR, toR, nCmdTimer, nPeriod );
        uint8_t G = interpolate( fromG, toG, nCmdTimer, nPeriod );
        uint8_t B = interpolate( fromB, toB, nCmdTimer, nPeriod );

        SetLights( R, G, B );
    }
}


/*

TODO at90usb162 only has 512bytes of sram.  
- Maybe keep working sample in eeprom.
- Customise VID/PID, device, etc
- inf - xp, vista
- tune leds when we have orb
- 3/4 leds

Samples 

Red
CFF0000

Colour cycle
C000000:2000:FF0000:2000:FFFF00:2000:00FF00:2000:00FFFF:2000:0000FF:2000:FF00FF:2000:FFFFFF:2000

Police Lights
CFF0000:10:FF0000:0:000000:100:000000:0:FF0000:10:FF0000:0:000000:100:000000:400:000000:0:0000FF:10:0000FF:0:000000:100:000000:0:0000FF:10:0000FF:0:000000:100:000000:0

Blink Red
CFF0000:500:FF0000:0:000000:500:000000:0
Blink Green
C00FF00:500:00FF00:0:000000:500:000000:0
Blink Blue
C0000FF:500:0000FF:0:000000:500:000000:0
Blink White
CFfffff:500:Ffffff:0:000000:500:000000:0
Blink Pink
CFf00ff:500:Ff00ff:0:000000:500:000000:0
Blink White
CFFFFFF:500:FFFFFF:0:000000:500:000000:0
White/Red
CFFFFFF:500:FFFFFF:0:000000:500:000000:0:FF0000:500:FF0000:0:000000:500:000000:0

*/
